/*
 * ----------------------------------------------
 *
 * This software belongs to the zzzutils package.
 *
 * Copyright (c) 2003 Fabian "zzznowman" Pietsch.
 * All rights reserved.
 *
 * ----------------------------------------------
 */

/*
 * zzzsearchdir.c -- search filesystem and give a prety-print file list
 *
 **********************************************************************
 *
 * brief ChangeLog:
 *   2003-08-06 to -14 C reimplementation of zzzupdate-xmms in progress
 *   2003-08-14 fork as zzzsearchdir
 *
 **********************************************************************
 *
 *
 */

#include "zzzutils.h"
#include "zbtree.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* required for stat() */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

/* required for readdir() */
#include <dirent.h>


/* init dirs count (cf. argv) */
int initdirsc = 1;

/* init dirs values (cf. argv) */
char *initdirsv[] = {
	"sto"
};

/* pointer to array of initial dirs
 * if program arguments present, points to first directory argument
 */
char **initdirsvp = initdirsv;

/* verbosity flag */
int verbose = 0;

/* search directory -- recursively search name and it's subdirectories */
int searchdir(const char *name)
{
	struct stat dir_info,		/* stat() for name,		*/
	            ent_info;		/*            dir entry		*/
	DIR *dir;			/* pointer to directory stream	*/
	struct dirent *dirent;		/* directory entry		*/
	struct zbtree *dir_dirs=NULL,	/* directory's subdirs,		*/
	              *dir_files=NULL,	/*             files		*/
		      *bt_cur;		/* zbtree enumeration		*/
	struct zbtree_enum *zbte;	/* enumerator for zbtree_next	*/
	char *cwd_old;			/* absolute path to current wd	*/
	char *name_new;			/* subdir (for recursion)	*/
	size_t name_len;		/* length of directory's path	*/


	/*
	 * check validity of the call
	 */
	
	/* stat name (for presence & directory check) */
	if (stat(name, &dir_info) != 0) {
		zperror("searchdir(): can't stat %s", name);
		return -1;
	}
	
	/* is name a directory? */
	if (!S_ISDIR(dir_info.st_mode)) {
		zerror("searchdir(): %s is not a directory", name);
		return -1;
	}


	/*
	 * do the real scan
	 */

	if (verbose > 0)
		printf("scanning %s...\n", name);
	
	/* open directory stream */
	if ((dir = opendir(name)) == NULL) {
		zperror("searchdir(): opendir(\"%s\")", name);
		return -1;
	}

	/* remember current directory to return there later */
	if ((cwd_old = getcwd(NULL, 0)) == NULL) {
		zperror("searchdir(): getcwd()");
		return -2;
	}
	
	/* enter directory */
	if (chdir(name) != 0) {
		zperror("searchdir(): couldn't chdir() to %s", name);
		free(cwd_old);
		closedir(dir);
		return -1;
	}
	
	/* loop through directory entries */
	while ((dirent = readdir(dir))) {
		/* current (.) or parent (..) directory? */
		if (dirent->d_name[0] == '.') {
			switch (dirent->d_name[1]) {
			case 0:
			case '.':
#ifdef DEBUG
				zwarn("searchdir(): rejecting %s", dirent->d_name);
#endif
				continue;
			}
		}
		
		/* stat directory entry */
		if (stat(dirent->d_name, &ent_info) != 0) {
			zperror("searchdir(): can't stat %s", dirent->d_name);
			continue;
		}
		
		if (S_ISDIR(ent_info.st_mode)) {
			zbtree_add(&dir_dirs, strdup(dirent->d_name));
#ifdef DEBUG
			zwarn("searchdir(): %s is a directory", dirent->d_name);
#endif
		}
		else {
			zbtree_add(&dir_files, strdup(dirent->d_name));
#ifdef DEBUG
			zwarn("searchdir(): %s is not a directory", dirent->d_name);
#endif
		}
	}
	dirent = NULL;

	/* leave directory */
	if (chdir(cwd_old) != 0) {
		zperror("searchdir(): couldn't chdir() back to %s", cwd_old);
		free(cwd_old);
		closedir(dir);
		zbtree_free(&dir_dirs, 1);
		zbtree_free(&dir_files, 1);
		return -2;
	}
	
	/* free absolute path to previous working directory */
	cwd_old = realloc(cwd_old, 0);
	
	/* close directory stream */
	if (closedir(dir) != 0)
		zperror("searchdir(): closedir()");


	/*
	 * process gathered information
	 */

	/* any files found? */
	if (dir_files) {
		printf("Files in %s:\n", name);

		/* start zbtree enumeration */
		if ((zbte = zbtree_estart(dir_files)) == NULL) {
			zerror("searchdir(): can't enumerate through zbtree dir_files");
			zbtree_free(&dir_files, 1);
			zbtree_free(&dir_dirs, 1);
			return -2;
		}

		/* walk through dir_files, automatically sorted by b-tree */
		while ((bt_cur = zbtree_enext(zbte)))
			printf("  %s\n", bt_cur->v);
		printf("\n");

		/* end zbtree enumeration */
		zbte = zbtree_eend(zbte);
	}

	/* any directories found? */
	if (dir_dirs) {
		name_len = strlen(name);
		
		/* start zbtree enumeration */
		if ((zbte = zbtree_estart(dir_dirs)) == NULL) {
			zerror("searchdir(): can't enumerate through zbtree dir_dirs");
			zbtree_free(&dir_files, 1);
			zbtree_free(&dir_dirs, 1);
			return -2;
		}

		/* walk through dir_dirs, automatically sorted by b-tree */
		while ((bt_cur = zbtree_enext(zbte))) {
#if defined(DEBUG)
			zwarn("searchdir(): descending to %s", bt_cur->v);
#endif
			/* allocate buffer for subdirectory's path */
			if ((name_new = calloc(name_len + strlen(bt_cur->v) + 2,
					       sizeof(char))) == NULL) {
				zperror("searchdir(): malloc()");
				zbtree_free(&dir_dirs, 1);
				zbtree_free(&dir_files, 1);
				return -2;
			}

			/* put together path: "thisdir/subdir" */
			strcpy(name_new, name);
			name_new[name_len] = '/';
			searchdir(strcat(name_new, bt_cur->v));

			/* free buffer for subdirectory's path */
			name_new = realloc(name_new, 0);
#if defined(DEBUG)
			zwarn("searchdir(): ascending from %s", bt_cur->v);
#endif
		}

		/* end zbtree enumeration */
		zbte = zbtree_eend(zbte);
	}

	/* free used zbtrees
	 * (including dirent.d_name string copies malloc()ed by strdup())
	 */
	zbtree_free(&dir_dirs, 1);
	zbtree_free(&dir_files, 1);
	
	return 0;
}

int main(const int argc, char *argv[])
{
	int initdir_idx;
	
	/* any arguments given? */
	if (argc > 1) {
		/* override default initial directories */
		initdirsvp = &argv[1];
		initdirsc = argc - 1;
	}

	/* do a searchdir() recursion for each initial directory */
	for (initdir_idx = 0; initdir_idx < initdirsc; initdir_idx++) {
		if (searchdir(initdirsvp[initdir_idx]) < -1) {
			zerror("searchdir() failed in a non-recoverable way -- exiting");
			exit(2);
		}
	}

	exit(0);
}
